/*
 * *************************************************
 * Copyright (c) 2018 QMTH. All Rights Reserved.
 * Created by Deason on 2018-08-27 16:24:08.
 * *************************************************
 */

package cn.com.qmth.examcloud.core.oe.admin.service.impl;

import cn.com.qmth.examcloud.api.commons.security.bean.User;
import cn.com.qmth.examcloud.core.basic.api.bean.CourseBean;
import cn.com.qmth.examcloud.core.oe.admin.base.jpa.Searcher;
import cn.com.qmth.examcloud.core.oe.admin.base.jpa.SpecUtils;
import cn.com.qmth.examcloud.core.oe.admin.base.jpa.SqlWrapper;
import cn.com.qmth.examcloud.core.oe.admin.base.utils.Check;
import cn.com.qmth.examcloud.core.oe.admin.base.utils.CommonUtil;
import cn.com.qmth.examcloud.core.oe.admin.dao.*;
import cn.com.qmth.examcloud.core.oe.admin.dao.entity.*;
import cn.com.qmth.examcloud.core.oe.admin.dao.enums.*;
import cn.com.qmth.examcloud.core.oe.admin.service.*;
import cn.com.qmth.examcloud.core.oe.admin.service.bean.examaudit.*;
import cn.com.qmth.examcloud.core.oe.admin.service.others.ExamCacheTransferHelper;
import cn.com.qmth.examcloud.examwork.api.bean.ExamBean;
import cn.com.qmth.examcloud.marking.api.MarkWorkCloudService;
import cn.com.qmth.examcloud.marking.api.request.AppendMarkWorkPaperReq;
import cn.com.qmth.examcloud.support.cache.bean.OrgCacheBean;
import cn.com.qmth.examcloud.support.enums.FaceBiopsyScheme;
import cn.com.qmth.examcloud.support.helper.FaceBiopsyHelper;
import cn.com.qmth.examcloud.web.helpers.GlobalHelper;
import com.google.common.collect.Lists;
import org.apache.commons.lang3.StringUtils;
import org.hibernate.query.NativeQuery;
import org.hibernate.transform.Transformers;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.persistence.EntityManager;
import javax.persistence.Query;
import java.math.BigInteger;
import java.util.Date;
import java.util.List;

import static cn.com.qmth.examcloud.core.oe.admin.service.bean.examaudit.ExamAuditMapper.EXAM_RECORD_DATA_ID;

/**
 * 考试记录审核相关接口
 *
 * @author: fengdesheng
 * @since: 2018/8/27
 */
@Service
public class ExamAuditServiceImpl implements ExamAuditService {
    @Autowired
    private ExamAuditRepo examAuditRepo;
    @Autowired
    private ExamScoreRepo examScoreRepo;
    @Autowired
    private EntityManager entityManager;

    @Autowired
    private ExamRecordDataRepo examRecordDataRepo;
    @Autowired
    private ExamRecordQuestionsService examRecordQuestionsService;
	@Autowired
	private ExamRecordForMarkingRepo examRecordForMarkingRepo;
    @Autowired
    private GainBaseDataService gainBaseDataService;
    @Autowired
    private ExamScorePushQueueService examScorePushQueueService;
    @Autowired
    private ExamScoreObtainQueueService examScoreObtainQueueService;
    @Autowired
    MarkWorkCloudService markWorkCloudService;
    @Autowired
    private ExamFaceLivenessVerifyRepo examFaceLivenessVerifyRepo;

    @Autowired
    private FaceBiopsyItemRepo faceBiopsyItemRepo;

    /**
     * 活体检测失败自动审核
     */
    private static final String FACE_LIVENESS_VERIFY_AUTO_AUDIT = "FACE_LIVENESS_VERIFY_AUTO_AUDIT";

    /**
     * 无照片自动审核
     */
    private static final String NOPHOTO_VERIFY_AUTO_AUDIT = "NOPHOTO_VERIFY_AUTO_AUDIT";

    /**
     * 系统审核
     */
    private static final String AUDIT_USER_NAME = "SYSTEM";
    @Override
    public Page<ExamAuditInfo> getExamAuditList(ExamAuditQuery query) {
        Check.isNull(query, "查询参数不能为空！");
        Check.isNull(query.getExamId(), "请先选择考试批次！");
        //封装查询条件
        Pageable pageable = SpecUtils.buildPageable(query.getPageNo(), query.getPageSize());
        final String columns = ExamAuditMapper.defaultColumns();
        SqlWrapper wrapper = new SqlWrapper()
                .select(columns).from("ec_oe_exam_audit").as("audit")
                .innerJoin("ec_oe_exam_record_data").as("record").on("record.id", "audit.exam_record_data_id")
                .where().notEq("record.exam_record_status", ExamRecordStatus.EXAM_INVALID.name());
        if (query.getOrgId() != null) {
            wrapper.and().eq("record.org_id", query.getOrgId());
        }
        if (query.getExamId() != null) {
            wrapper.and().eq("record.exam_id", query.getExamId());
        }
        if (StringUtils.isNotBlank(query.getStudentCode())) {
            wrapper.and().like("record.student_code", query.getStudentCode());
        }
        if (StringUtils.isNotBlank(query.getStudentName())) {
            wrapper.and().like("record.student_name", query.getStudentName());
        }
        if (StringUtils.isNotBlank(query.getIdentityNumber())) {
            wrapper.and().like("record.identity_number", query.getIdentityNumber());
        }
        if (query.getCourseId() != null) {
            wrapper.and().eq("record.course_id", query.getCourseId());
        }
        if (StringUtils.isNotBlank(query.getCourseLevel())) {
            wrapper.and().eq("record.course_level", query.getCourseLevel());
        }
        if (query.getExamRecordDataId() != null) {
            wrapper.and().eq("record.id", query.getExamRecordDataId());
        }
        if (StringUtils.isNotBlank(query.getStatus())) {
            wrapper.and().eq("audit.status", query.getStatus());
        }
        if(StringUtils.isNoneBlank(query.getDisciplineType())){
        	wrapper.and().eq("audit.discipline_type", query.getDisciplineType());
        }
        if(StringUtils.isNoneBlank(query.getAuditUserName())){
        	wrapper.and().like("audit.audit_user_name", query.getAuditUserName());
        }
        long totalSize = 0;
        //查询总记录数
        if(query.getSelectType() == null || query.getSelectType() != SelectType.EXPORT){
        	final String count = "count(0)";
            String countSql = wrapper.build().replace(columns, count);
            Query countQuery = entityManager.createNativeQuery(countSql);
            BigInteger element = (BigInteger) countQuery.getSingleResult();
            totalSize = element.longValue();
            if (totalSize == 0) {
                return new PageImpl<>(Lists.newArrayList(), pageable, totalSize);
            }
        }

        //查询分页记录
        wrapper.orderBy("record.id", true);
        Query dataQuery = entityManager.createNativeQuery(wrapper.build());

//        dataQuery.unwrap(SQLQuery.class).setResultTransformer(Transformers.aliasToBean(HashMap.class));
        dataQuery.unwrap(NativeQuery.class).setResultTransformer(Transformers.ALIAS_TO_ENTITY_MAP);
        if(query.getSelectType() == null || query.getSelectType() != SelectType.EXPORT){
	        dataQuery.setFirstResult((int) pageable.getOffset());
	        dataQuery.setMaxResults(pageable.getPageSize());
        }
        List resultList = dataQuery.getResultList();

        List<ExamAuditInfo> list = ExamAuditEntityConvert.of(resultList);
        
        for(ExamAuditInfo examAuditInfo:list){
            OrgCacheBean orgBean = gainBaseDataService.getOrgBean(examAuditInfo.getOrgId());
            examAuditInfo.setOrgName(orgBean.getName());

            CourseBean courseBean = ExamCacheTransferHelper.getCachedCourse(examAuditInfo.getCourseId());
        	examAuditInfo.setCourseName(courseBean.getName());
        	examAuditInfo.setCourseCode(courseBean.getCode());
        	examAuditInfo.setCourseLevel(CourseLevel.getCourseLevel(courseBean.getLevel()).getTitle());

            //获取考试名称
            ExamBean examBean = ExamCacheTransferHelper.getCachedExam(query.getExamId(),examAuditInfo.getStudentId());
            examAuditInfo.setExamName(examBean.getName());

        }
        return new PageImpl<ExamAuditInfo>(list, pageable, totalSize);
    }

    @Override
    public ExamAuditEntity getExamAuditByExamRecordDataId(Long examRecordDataId) {
        if (examRecordDataId == null) {
            return null;
        }
        Searcher searchers = new Searcher()
                .eq(EXAM_RECORD_DATA_ID, examRecordDataId);
        Specification<ExamAuditEntity> spec = SpecUtils.buildSearchers(ExamAuditEntity.class, searchers.build());
        List<ExamAuditEntity> list = examAuditRepo.findAll(spec);
        if (list != null && list.size() > 0) {
            return list.get(0);
        }
        return null;
    }
    
    @Override
    public void batchAudit(List<Long> examRecordDataIds, Boolean isPass,User user) {
        for (Long examRecordDataId : examRecordDataIds) {
        	DisciplineType disciplineType = isPass?DisciplineType.BATCH_PASS:DisciplineType.BATCH_NOTPASS;
        	singleAudit(examRecordDataId,isPass,"批量审核",disciplineType,user);
        }
    }

	@Override
	@Transactional
	public void singleAudit(Long examRecordDataId, Boolean isPass,String disciplineDetail, DisciplineType disciplineType,User user) {
		ExamRecordDataEntity examRecordData = GlobalHelper.getEntity(examRecordDataRepo,examRecordDataId,ExamRecordDataEntity.class);
        if (examRecordData == null) {
            return;
        }
        //已经审核过了，这里直接返回，如果需要重新审核，调用重审redoAudit方法
        if(examRecordData.getIsAudit() !=null && examRecordData.getIsAudit()){
        	return;
        }
        
        //保存审核记录
        ExamAuditEntity examAudit = new ExamAuditEntity();
        examAudit.setExamRecordDataId(examRecordDataId);
        examAudit.setStatus(isPass ? AuditStatus.PASS : AuditStatus.UN_PASS);
        examAudit.setDisciplineType(disciplineType);
        examAudit.setDisciplineDetail(disciplineDetail);
        if(user != null){
        	examAudit.setUserId(user.getUserId().toString());
        	examAudit.setAuditUserName(user.getDisplayName());
        }
        examAuditRepo.save(examAudit);
        
        //保存考试记录
        examRecordData.setIsAudit(true);
        examRecordData.setIsIllegality(isPass?false:true);
        examRecordDataRepo.save(examRecordData);
        
        doAuditPostProcesser(isPass,examRecordData);
	}
	
	@Override
	public void redoAudit(RedoAuditInfo redoAuditInfo,User user) {
		for(Long examRecordDataId:redoAuditInfo.getExamRecordDataIds()){
			ExamAuditEntity examAudit = examAuditRepo.findByExamRecordDataId(examRecordDataId);
			if(examAudit == null){
				examAudit = new ExamAuditEntity();
				examAudit.setCreationTime(new Date());
				examAudit.setExamRecordDataId(examRecordDataId);
			}
			examAudit.setAuditUserName(user.getDisplayName());
			examAudit.setUserId(user.getUserId().toString());
			if(StringUtils.isBlank(redoAuditInfo.getDisciplineDetail())){
				examAudit.setDisciplineDetail("重审");
			}else{
				examAudit.setDisciplineDetail(redoAuditInfo.getDisciplineDetail());
			}
			examAudit.setDisciplineType(DisciplineType.tramsformByType(redoAuditInfo.getDisciplineType()));
			examAudit.setStatus(redoAuditInfo.getIsPass()?AuditStatus.PASS:AuditStatus.UN_PASS);
			examAudit.setUpdateTime(new Date());
			examAuditRepo.save(examAudit);
			
			ExamRecordDataEntity examRecordData = GlobalHelper.getEntity(examRecordDataRepo,examRecordDataId,ExamRecordDataEntity.class);
			examRecordData.setIsWarn(true);
			examRecordData.setIsAudit(true);
			examRecordData.setIsIllegality(!redoAuditInfo.getIsPass());//是否违纪
			examRecordDataRepo.save(examRecordData);
			
			doAuditPostProcesser(redoAuditInfo.getIsPass(),examRecordData);
		}
	}

    /**
     * 活体检测失败，系统审核
     *
     * @param examRecordDataId
     * @param rootOrgId
     */
    @Override
    public void saveExamAuditByFaceVerifyFailed(Long examRecordDataId, Long rootOrgId) {
        ExamAuditEntity examAudit = examAuditRepo.findByExamRecordDataId(examRecordDataId);
        if (examAudit != null) {
            return;
        }
        ExamAuditEntity examAuditEntity = new ExamAuditEntity();
        examAuditEntity.setExamRecordDataId(examRecordDataId);
        examAuditEntity.setDisciplineType(DisciplineType.ACTION_FAILURE);
        examAuditEntity.setUserId(FACE_LIVENESS_VERIFY_AUTO_AUDIT);//活体检测自动审核
        examAuditEntity.setAuditUserName(AUDIT_USER_NAME);
        examAuditEntity.setDisciplineDetail(getFaceVerifyDisciplineDetail(examRecordDataId, rootOrgId));//计算违纪描述
        examAuditEntity.setStatus(AuditStatus.UN_PASS);
        examAuditRepo.save(examAuditEntity);
    }

    /**
     * 抓拍照片为0，系统审核
     *
     * @param examRecordDataId
     */
    @Override
    public void saveExamAuditByNoPhoto(Long examRecordDataId) {
        ExamAuditEntity examAudit = examAuditRepo.findByExamRecordDataId(examRecordDataId);
        if (examAudit != null) {
            return;
        }
        ExamAuditEntity examAuditEntity = new ExamAuditEntity();
        examAuditEntity.setExamRecordDataId(examRecordDataId);
        examAuditEntity.setDisciplineType(DisciplineType.OTHER);
        examAuditEntity.setDisciplineDetail("抓拍照片数量为0");
        examAuditEntity.setUserId(NOPHOTO_VERIFY_AUTO_AUDIT);//抓拍照片为0自动审核
        examAuditEntity.setAuditUserName(AUDIT_USER_NAME);
        examAuditEntity.setStatus(AuditStatus.UN_PASS);
        examAuditRepo.save(examAuditEntity);
    }

    //获取活检违纪详情
    private String getFaceVerifyDisciplineDetail(Long examRecordDataId, Long rootOrgId) {
        FaceBiopsyScheme faceBiopsyScheme = FaceBiopsyHelper.getFaceBiopsyScheme(rootOrgId);

        //新活体检测方案从新表中取相关数据
        if (faceBiopsyScheme == FaceBiopsyScheme.NEW) {
            List<FaceBiopsyItemEntity> faceBiopsyItems = faceBiopsyItemRepo.findByExamRecordDataIdOrderByIdAsc(examRecordDataId);
            //从未调过人脸检测接口，或者只调过接口但是未保存过结果(即未完成活检)，则提示未进行人脸活体检测
            if (faceBiopsyItems == null ||
                    (faceBiopsyItems.size() == 1 && !faceBiopsyItems.get(0).getCompleted())) {
                return "未进行人脸活体检测";
            }

            StringBuffer sbBuffer = new StringBuffer("");
            for (FaceBiopsyItemEntity item : faceBiopsyItems) {
                String startTime = CommonUtil.getDateStrWithSecond(item.getCreationTime());
                String resultMsg = "";
                if (null != item.getCompleted() && !item.getCompleted()) {
                    resultMsg = "活检未完成";
                } else {
                    if (null != item.getResult() && item.getResult()) {
                        resultMsg = "活检成功";
                    } else {
                        resultMsg = item.getErrorMsg() == null ? "" : item.getErrorMsg();
                    }
                }
                sbBuffer.append(startTime + ":" + resultMsg);
                sbBuffer.append("&&");
            }
            return sbBuffer.toString();
        }
        //旧活体检测方案从旧表中取相关数据
        else {
            List<ExamFaceLivenessVerifyEntity> faceVerifies =
                    examFaceLivenessVerifyRepo.findByExamRecordDataIdOrderById(examRecordDataId);
            if (faceVerifies == null || faceVerifies.size() == 0) {
                return "未进行人脸动作检测";
            } else {
                StringBuffer sbBuffer = new StringBuffer("");
                for (ExamFaceLivenessVerifyEntity faceVerify : faceVerifies) {
                    String startTime = CommonUtil.getDateStrWithSecond(faceVerify.getStartTime());
                    sbBuffer.append(startTime + ":" + FaceVerifyResult.getDescByName(faceVerify.getVerifyResult()));
                    sbBuffer.append("&&");
                }
                String disciplineDetail = sbBuffer.toString();
                if (disciplineDetail.length() > 500) {
                    disciplineDetail = disciplineDetail.substring(0, 500);
                }
                return disciplineDetail;
            }
        }
    }

    /**
	 * 重审后置处理
	 * @param isPass
	 * @param examRecordData
	 */
	private void doAuditPostProcesser(Boolean isPass,ExamRecordDataEntity examRecordData) {
		if(isPass){
			auditPassPostProcesser(examRecordData);
		}else{
			auditUnPassPostProcesser(examRecordData);
		}
	}
	
	/**
	 * 审核通过后置处理
	 */
	private void auditPassPostProcesser(ExamRecordDataEntity examRecordData){
    	ExamScoreEntity examScore = examScoreRepo.findByExamRecordDataId(examRecordData.getId());
    	//审核通过且为非全客观题卷，保存阅卷需要数据
    	saveExamRecordForMarkingWithAudit(examRecordData,examScore);
    	//保存推送分数队列
        examScorePushQueueService.saveScoreDataInfoToQueue(examRecordData.getId(),examScore.getId());
        //保存获取分数队列
        examScoreObtainQueueService.saveExamScoreObtainQueue(examRecordData.getId());
        //发送通知
        examScoreObtainQueueService.sendObtainScoreNotify(examRecordData.getRootOrgId());
	}

    /**
     * 审核通过且为非全客观题卷，保存阅卷需要数据
     * @param examRecordData
     */
	private void saveExamRecordForMarkingWithAudit(ExamRecordDataEntity examRecordData,ExamScoreEntity examScore){
    	if(examRecordData.getIsAllObjectivePaper() != null && !examRecordData.getIsAllObjectivePaper()){
            ExamRecordForMarkingEntity examRecordForMarkingExists =
                    examRecordForMarkingRepo.findByExamRecordDataId(examRecordData.getId());
            if (examRecordForMarkingExists != null) {
                return;
            }
            ExamRecordForMarkingEntity examRecordForMarking = new ExamRecordForMarkingEntity();
            examRecordForMarking.setExamId(examRecordData.getExamId());
            examRecordForMarking.setExamRecordDataId(examRecordData.getId());
            examRecordForMarking.setExamStudentId(examRecordData.getExamStudentId());
            examRecordForMarking.setBasePaperId(examRecordData.getBasePaperId());
            examRecordForMarking.setPaperType(examRecordData.getPaperType());
            examRecordForMarking.setCourseId(examRecordData.getCourseId());
            examRecordForMarking.setObjectiveScore(examScore.getObjectiveScore());
            int subjectiveAnswerLength = examRecordQuestionsService.calculationSubjectiveAnswerLength(examRecordData.getId());
            examRecordForMarking.setSubjectiveAnswerLength(subjectiveAnswerLength);
            examRecordForMarkingRepo.save(examRecordForMarking);

    		//调用阅卷接口，通知有新数据
            AppendMarkWorkPaperReq appendMarkWorkPaperReq = new AppendMarkWorkPaperReq();
            appendMarkWorkPaperReq.setRootOrgId(examRecordData.getRootOrgId());
            appendMarkWorkPaperReq.setExamId(examRecordData.getExamId());
            markWorkCloudService.appendMarkWorkPaper(appendMarkWorkPaperReq);
    	}
	}

	/**
	 * 审核不通过后置处理
	 * @param examRecordData
	 */
	private void auditUnPassPostProcesser(ExamRecordDataEntity examRecordData) {
		ExamRecordForMarkingEntity examRecordForMarking = examRecordForMarkingRepo.findByExamRecordDataId(examRecordData.getId());
		if(examRecordForMarking != null){
			//删除阅卷需要数据，防止违纪数据进入阅卷
			examRecordForMarkingRepo.delete(examRecordForMarking);
		}
		//保存获取分数队列
        examScoreObtainQueueService.saveExamScoreObtainQueue(examRecordData.getId());
        //发送通知
        examScoreObtainQueueService.sendObtainScoreNotify(examRecordData.getRootOrgId());
	}
}